/*/../bin/true; p="/tmp/$(basename "$0").elf"; v=1; gcc -Wfatal-errors -std=c99 -O42 "$0" -o "$p" && chmod 755 "$p" && ("$p"; v=$?; rm "$p"); exit $v;*/


/*
 * yprof - y as in why
 * v1.0, 2020-02-27, ed <irc.rizon.net>, MIT-licensed
 * https://ocv.me/dev/?yprof.c
 */


// config: number of lines to show in report
#define YPROF_TRUNC 5


// config: perform cpu time sampling for the process;
//         either leave undefined -OR- getpid -OR- getppid
#define YPROF_CPU getpid


// convenience: take a sample
#define YP yprof_put(wv, __LINE__)


// ok here we go
//


#define _POSIX_C_SOURCE 199309L

#include <stdlib.h>
#include <stdint.h>
#include <inttypes.h>
#include <stdio.h>
#include <time.h>


#ifdef YPROF_CPU
#include <unistd.h>
#endif


typedef struct yprof_sample {
	time_t sec;
	long nsec;
#ifdef YPROF_CPU
	long cpu_usr;
	long cpu_sys;
#endif
	long lineno;
} yprof_sample;


typedef struct yprof_vec {
	size_t size;
	size_t ubound;
	yprof_sample** data;
} yprof_vec;


#ifdef YPROF_CPU
int yprof_getcpu(long* cpu_usr, long* cpu_sys) {
	char fn[32];
	const long me = YPROF_CPU();
	sprintf(fn, "/proc/%ld/stat", me);
	FILE* fh = fopen(fn, "rb");
	if (!fh)
		return -1;
	
	char buf[1024];
	size_t nread = fread(buf, 1, sizeof buf, fh);
	
	char* ptr = buf + 32;
	while (*--ptr != ')')
		if (ptr < buf)
			return -1;
	
	if (*ptr != ')')
		return -1;
	
	ptr += 4;
	for (int n = 4; n < 14; n++)
		while (*ptr++ != ' ')
			if (*ptr == '\0')
				return -1;
	
	char* ptr2 = ptr;
	while (*ptr2++ != ' ')
		if (*ptr2 == '\0')
			return -1;
	
	char* ptr3 = ptr2;
	while (*ptr3++ != ' ')
		if (*ptr3 == '\0')
			return -1;
	
	*cpu_usr = strtol(ptr, &ptr2, 10);
	*cpu_sys = strtol(ptr2, &ptr3, 10);
	fclose(fh);
	return 0;
}
#endif


int yprof_cmp(const void* a, const void* b) {
	const struct yprof_sample* sa = *(const struct yprof_sample**)a;
	const struct yprof_sample* sb = *(const struct yprof_sample**)b;
	//printf("[%lx] [%lx] [%ld] [%ld] [%u] [%u]\n", a, b, sa->nsec, sb->nsec, sa->lineno, sb->lineno);
	
	const int64_t ret =
		sa->sec != sb->sec ?
			sa->sec - sb->sec :
			sa->nsec - sb->nsec;
	
	//return ret < 0 ? 1 : ret > 0 ? -1 : 0;
	if (ret < 0) return 1;
	if (ret > 0) return -1;
	return 0;
}


struct yprof_vec* yprof_new(size_t sz) {
	struct yprof_vec* wv = malloc(sizeof(struct yprof_vec));
	wv->size = 0;
	wv->ubound = sz;
	wv->data = malloc(wv->ubound * sizeof(struct yprof_sample));
	if (wv->data)
		return wv;
	
	return 0;
}


void yprof_del(struct yprof_vec* wv) {
	for (size_t a = 0; a < wv->size; ++a)
		free(wv->data[a]);
	
	free(wv->data);
	free(wv);
}


int yprof_put(struct yprof_vec* wv, long lineno) {
	struct timespec tp;
	clock_gettime(CLOCK_MONOTONIC, &tp);
	
	if (wv->size >= wv->ubound) {
		wv->ubound *= 2;
		wv->data = realloc(wv->data, wv->ubound * sizeof(struct yprof_sample));
		if (!wv->data)
			return -1;
	}
	
	struct yprof_sample* s = malloc(sizeof(struct yprof_sample));
	
	s->sec = tp.tv_sec;
	s->nsec = tp.tv_nsec;
	s->lineno = lineno;

#ifdef YPROF_CPU
	yprof_getcpu(&s->cpu_usr, &s->cpu_sys);
#endif
	
	wv->data[wv->size++] = s;
	return 0;
}


void yprof_report(struct yprof_vec* wv) {
	time_t sec0 = wv->data[0]->sec;
	long nsec0 = wv->data[0]->nsec;
#ifdef YPROF_CPU
	long cpu_usr0 = wv->data[0]->cpu_usr;
	long cpu_sys0 = wv->data[0]->cpu_sys;
#endif
	
	for (size_t a = wv->size - 1; a > 0; --a) {
		const struct yprof_sample* s0 = wv->data[a-1];
		struct yprof_sample* s = wv->data[a];
		
		if (s->nsec >= s0->nsec) {
			s->sec -= s0->sec;
			s->nsec -= s0->nsec;
		}
		else {
			s->nsec = s0->nsec - s->nsec;
			s->sec = (s->sec - s0->sec) - 1;
		}
#ifdef YPROF_CPU
		s->cpu_usr -= s0->cpu_usr;
		s->cpu_sys -= s0->cpu_sys;
#endif
	}
	wv->data[0]->sec = 0;
	wv->data[0]->nsec = 0;
#ifdef YPROF_CPU
	wv->data[0]->cpu_usr = 0;
	wv->data[0]->cpu_sys = 0;
#endif
	
	qsort(wv->data, wv->size, sizeof(struct yprof_sample*), yprof_cmp);
	
#ifdef YPROF_CPU
	fprintf(stdout, "%ld'%ld Lx %ld %ld\n", sec0, nsec0, cpu_usr0, cpu_sys0);
#else
	fprintf(stdout, "%ld'%ld\n", sec0, nsec0);
#endif

	const size_t end = wv->size > YPROF_TRUNC ? YPROF_TRUNC : wv->size;
	for (size_t a = 0; a < end; ++a) {
		const yprof_sample s = *wv->data[a];
#ifdef YPROF_CPU
		fprintf(stdout, "%12lld L%ld %ld %ld\n",
			s.sec * 1000000000LL + s.nsec, s.lineno, s.cpu_usr, s.cpu_sys);
#else
		fprintf(stdout, "%12lld L%ld\n", s.sec * 1000000000LL + s.nsec, s.lineno);
#endif
		// stderr = ~260% cost
	}
	yprof_del(wv);
}


void test_yprof_1() {
	yprof_vec *wv = yprof_new(24);
	YP;
	YP;
	YP; int a = 3;
	YP;
	YP;
	YP; a += 2;
	YP;
	YP;
	YP;
	if (a > 3) {
		YP;
		YP;
		YP;
	}
	YP;
	YP;
	YP;
	if (a < 3) {
		YP;
		YP;
		YP;
	}
	YP;
	YP;
	YP; a -= 2;
	YP;
	YP;
	YP; yprof_report(wv);
	//yprof_del(wv);
}


void test_yprof_2() {
	yprof_vec *wv = yprof_new(24);
	YP;
	YP; nanosleep((const struct timespec[]){{0, 1000000L}}, NULL);
	YP; nanosleep((const struct timespec[]){{0, 2000000L}}, NULL);
	YP; nanosleep((const struct timespec[]){{0, 3000000L}}, NULL);
	YP; nanosleep((const struct timespec[]){{0, 4000000L}}, NULL);
	YP; nanosleep((const struct timespec[]){{0, 5000000L}}, NULL);
	YP; nanosleep((const struct timespec[]){{0, 6000000L}}, NULL);
	YP; nanosleep((const struct timespec[]){{0, 7000000L}}, NULL);
	YP; nanosleep((const struct timespec[]){{0, 8000000L}}, NULL);
	YP; nanosleep((const struct timespec[]){{0, 9000000L}}, NULL);
	YP; nanosleep((const struct timespec[]){{0, 10000000L}}, NULL);
	YP; nanosleep((const struct timespec[]){{0, 11000000L}}, NULL);
	YP; nanosleep((const struct timespec[]){{0, 12000000L}}, NULL);
	YP; nanosleep((const struct timespec[]){{0, 13000000L}}, NULL);
	YP; nanosleep((const struct timespec[]){{0, 14000000L}}, NULL);
	YP; nanosleep((const struct timespec[]){{0, 15000000L}}, NULL);
	YP; nanosleep((const struct timespec[]){{0, 16000000L}}, NULL);
	YP; nanosleep((const struct timespec[]){{0, 17000000L}}, NULL);
	YP; nanosleep((const struct timespec[]){{0, 18000000L}}, NULL);
	YP; nanosleep((const struct timespec[]){{0, 19000000L}}, NULL);
	YP; yprof_report(wv);
	//  3295547 L263 0 0
	//  2933208 L262 0 0
	//  1060611 L261 0 0
	//    17229 L259 0 0
	//     3880 L260 0 0
}


void test_yprof_1_loop() {
	for (size_t a = 0; a < 500000; ++a)
		test_yprof_1();
	
	// wall 1.885   user 1.834   sys 0.046   500000 without cpu
	// wall 3.306   user 0.985   sys 2.251    50000 with cpu
}


#ifdef YPROF_CPU
void test_getcpu_1() {
	for (;;) {
		long cpu_usr, cpu_sys;
		for (int a = 0; a < 10000; a++)
			yprof_getcpu(&cpu_usr, &cpu_sys);
		
		printf("[%ld] [%ld]\n", cpu_usr, cpu_sys);
		
		for (int a = 0; a < 300000; a++)
			yprof_del(yprof_new(12));
		
		for (int a = 0; a < 1000; a++)
			nanosleep((const struct timespec[]){{0, 10L}}, NULL);
		
		continue;
	}
}


void test_getcpu_2() {
	long x1, x2;
	for (int a = 0; a < 500000; a++)
		yprof_getcpu(&x1, &x2);
	
	// wall 1.487   user 0.326   sys 1.123
}
#endif


int main(int argc, char** argv) {
	//test_yprof_1_loop();
	//test_yprof_1();
	test_yprof_2();
	//test_getcpu_1();
	//test_getcpu_2();
	return 0;
}


/* test stuff

TIMEFORMAT="wall %3R   user %3U   sys %3S"; gcc -Wfatal-errors -std=c99 -O42 yprof.c && for ((n=0; n<10; n++)); do time ./a.out > /dev/null; done

while true; do t=$(stat -c%Y yprof.c); sleep 0.2; [ $t = "$ot" ] && continue; ot="$t"; printf '\n\n\n\n\n'; gcc -Wfatal-errors -std=c99 -O0 yprof.c && valgrind --leak-check=full --track-origins=yes --partial-loads-ok=no --expensive-definedness-checks=yes --exit-on-first-error=yes --error-exitcode=1 ./a.out; done

*/
